#ifndef __BEOS__
#ifndef __MORPHOS__

#include "../stdafx.h"
#include "ttdstring.h"

#include "../openttd.h"
#include "../hal.h"
#include "../sound.h"
#include "../string.h"
#include "fcntl.h"
#include <wtypes.h>
#include "extmidi.h"
#include "../../TimidityCE\config.h"
#include "../../TimidityCE\common.h"
#include "../../TimidityCE\instrum.h"
#include "../../TimidityCE\playmidi.h"
#include "../../TimidityCE\readmidi.h"
#include "../../TimidityCE\output.h"
#include "../../TimidityCE\tables.h"
#include "../../TimidityCE\myprintf.h"
#include "../../TimidityCE\errno.h"

#define MAX_WIDE_STR	1024
#define MULTI_TO_WIDE( multi, wide, lenWide )  MultiByteToWideChar( CP_ACP,MB_PRECOMPOSED, multi, strlen(multi) + 1, wide, lenWide)
#define MAX_MB_STR	1024
#define WIDE_TO_MULTI( wide, multi, lenMulti) WideCharToMultiByte( CP_ACP, 0, wide, _tcslen(wide) + 1, multi, lenMulti, NULL, NULL )

char song[MAX_PATH];
DWORD threadId;

int free_instruments_afterwards = 1;
// MYCODE EDIT
static wchar_t def_instr_name[256] = TEXT ("Patch\\acpiano.pat");
//ControlMode ctl;
extern MidiEvent *current_event;

MIDICONTROL _midi;

CRITICAL_SECTION critSect;

#define MAXWORDS 10

_TCHAR *GetPathWide();

_TCHAR *GetPathWide()
{
        static _TCHAR lpFilename[MAX_PATH + 1]; // fix memory leak
        _TCHAR *pDest; 
        int nStrLen; 
        int result; 
        int ch = '\\'; 
		static char strMB[MAX_PATH + 1];
        nStrLen = GetModuleFileName(NULL, lpFilename, MAX_PATH + 1); 

        pDest = wcsrchr(lpFilename, ch); 
		
        if (pDest != NULL) 
        { 
			result = pDest - lpFilename + 1;
			lpFilename[result] = '\0';
			return lpFilename; 
        } 
        else 
        {
			return NULL; 
        }
} 

static int
read_config_file (BSTR name)
{
	BSTR error=TEXT("Error!");
	FILE *fp;
	wchar_t tmp[1024], *w[MAXWORDS], *cp;
	char buffer[1024];
	ToneBank *bank = 0;
	int i, j, k, line = 0, words;
	static int rcf_count = 0;
	_TCHAR strWide[MAX_WIDE_STR];

	if (rcf_count > 50) {
		myprintf (TEXT("Error!"),TEXT("Probable source loop in configuration files"));
		return (-1);
    }
	
	_stprintf(strWide, TEXT("%s%s"), GetPathWide(), name);

	if (!(fp = open_file (strWide, 1, OF_VERBOSE)))
		return -1;
	
	while (fgets (buffer, 1024, fp)) {
		memset(tmp,0,1024*2);
		MultiByteToWideChar(CP_ACP,0,buffer,-1,tmp,1024);
		line++;
		w[words = 0] = wcstok (tmp, TEXT (" \t\r\n\240"));
		if (!w[0] || (*w[0] == '#')) {
			continue;
		}
		while (w[words] && (words < MAXWORDS)) {
			w[++words] = wcstok (0, TEXT (" \t\r\n\240"));
		}
		if (!wcscmp (w[0], TEXT ("dir"))) {
			if (words < 2) {
				myprintf (error,L"%s: line %d: No directory given\n", name, line);
				return -2;
			}
			for (i = 1; i < words; i++)
				add_to_pathlist (w[i]);
		} else if (!wcscmp (w[0], TEXT ("source"))) {
			if (words < 2) {
				myprintf (error, L"%s: line %d: No file name given\n", name, line);
				return -2;
			}
			for (i = 1; i < words; i++) 
			{
				rcf_count++;
				_stprintf(strWide, TEXT("%s%s"), GetPathWide(), w[i]);
				read_config_file (strWide);
				rcf_count--;
			}
		} else if (!wcscmp (w[0], TEXT ("default"))) {
			if (words != 2)	{
				myprintf(error,L"%s: line %d: Must specify exactly one patch name\n",
					name, line);
				return -2;
			}
			wcsncpy (def_instr_name, w[1], 255);
			def_instr_name[255] = '\0';
		} else if (!wcscmp (w[0], TEXT ("drumset"))) {
			if (words < 2) {
				myprintf(error,L"%s: line %d: No drum set number given\n",
					name, line);
				return -2;
			}
			i = _wtoi (w[1]);
			if (i < 0 || i > 127) {
				myprintf (error,
					L"%s: line %d: Drum set must be between 0 and 127\n",
					name, line);
				return -2;
			}
			if (!drumset[i]) {
				drumset[i] = (ToneBank *)safe_malloc (sizeof (ToneBank));
				memset (drumset[i], 0, sizeof (ToneBank));
			}
			bank = drumset[i];
		}
		else if (!wcscmp (w[0], TEXT ("bank"))) {
			if (words < 2) {
				myprintf(error,L"%s: line %d: No bank number given\n",
					name, line);
				return -2;
			}
			i = _wtoi (w[1]);
			if (i < 0 || i > 127) {
				myprintf(error,
					L"%s: line %d: Tone bank must be between 0 and 127\n",
					name, line);
				return -2;
			}
			if (!tonebank[i]) {
				tonebank[i] = (ToneBank *)safe_malloc (sizeof (ToneBank));
				memset (tonebank[i], 0, sizeof (ToneBank));
			}
			bank = tonebank[i];
		} else {
			if ((words < 2) || (*w[0] < '0' || *w[0] > '9')) {
				myprintf (error,L"%s: line %d: syntax error\n", name, line);
				return -2;
			}
			i = _wtoi (w[0]);
			if (i < 0 || i > 127) {
				myprintf (error,
					L"%s: line %d: Program must be between 0 and 127\n",
					name, line);
				return -2;
			}
			if (!bank) {
				myprintf (error,
					L"%s: line %d: Must specify tone bank or drum set "
					L"before assignment\n", name, line);
				return -2;
			}
			if (bank->tone[i].name) {
				free (bank->tone[i].name);
			}
			wcscpy ((bank->tone[i].name = (BSTR)safe_malloc(2*(wcslen(w[1])+1))),w[1]);
			bank->tone[i].note = bank->tone[i].amp = bank->tone[i].pan =
				bank->tone[i].strip_loop = bank->tone[i].strip_envelope =
				bank->tone[i].strip_tail = -1;
			
			for (j = 2; j < words; j++) {
				if (!(cp = wcschr (w[j], '='))) {
					myprintf (error, L"%s: line %d: bad patch option %s\n",
						name, line, w[j]);
					return -2;
				}
				*cp++ = 0;
				if (!wcscmp (w[j], TEXT ("amp"))) {
					k = _wtoi (cp);
					if ((k < 0 || k > MAX_AMPLIFICATION)
						|| (*cp < '0' || *cp > '9')) {
						myprintf (error,
							L"%s: line %d: amplification must be between "
							L"0 and %d\n", name, line, MAX_AMPLIFICATION);
						return -2;
					}
					bank->tone[i].amp = k;
				} else if (!wcscmp (w[j], TEXT ("note"))) {
					k = _wtoi (cp);
					if ((k < 0 || k > 127) || (*cp < '0' || *cp > '9'))	{
						myprintf (error,
							L"%s: line %d: note must be between 0 and 127\n",
							name, line);
						return -2;
					}
					bank->tone[i].note = k;
				} else if (!wcscmp (w[j], TEXT ("pan"))) {
					if (!wcscmp (cp, TEXT ("center")))
						k = 64;
					else if (!wcscmp (cp, TEXT ("left")))
						k = 0;
					else if (!wcscmp (cp, TEXT ("right")))
						k = 127;
					else
						k = ((_wtoi (cp) + 100) * 100) / 157;
					if ((k < 0 || k > 127) ||
						(k == 0 && *cp != '-' && (*cp < '0' || *cp > '9')))	{
						myprintf (error,
							L"%s: line %d: panning must be left, right, "
							L"center, or between -100 and 100\n",
							name, line);
						return -2;
					}
					bank->tone[i].pan = k;
				} else if (!wcscmp (w[j], TEXT ("keep"))) {
					if (!wcscmp (cp, TEXT ("env")))
						bank->tone[i].strip_envelope = 0;
					else if (!wcscmp (cp, TEXT ("loop")))
						bank->tone[i].strip_loop = 0;
					else {
						myprintf (error,
							L"%s: line %d: keep must be env or loop\n",
							name, line);
						return -2;
					}
				} else if (!wcscmp (w[j], TEXT ("strip"))) {
					if (!wcscmp (cp, TEXT ("env")))
						bank->tone[i].strip_envelope = 1;
					else if (!wcscmp (cp, TEXT ("loop")))
						bank->tone[i].strip_loop = 1;
					else if (!wcscmp (cp, TEXT ("tail")))
						bank->tone[i].strip_tail = 1;
					else {
						myprintf (error,
							L"%s: line %d: strip must be env, loop, or tail\n",
							name, line);
						return -2;
					}
				} else {
					myprintf(error, L"%s: line %d: bad patch option %s\n",
						name, line, w[j]);
					return -2;
				}
			}
		}
    }
	if (ferror (fp)) {
		BSTR text;
		
		text = getError(errno);
		
		myprintf(error, L"Can't read %s: %s\n", name, text);
		close_file (fp);
		LocalFree(text);
		return -2;
    }
	close_file (fp);
	return 0;
}

extern "C" void ExtMidiStop(void)
{
	if (_midi.playing) 
	{
		_midi.stop_song = true;
		_midi.start_song[0] = '\0';
	}
	
	_midi.terminate = true;
	CloseHandle(&threadId);
	//ctl.isPlaying = false;
	play_mode->close_output();
	//ctl.close();
	DeleteCriticalSection(&critSect);
}

extern "C" void ExtMidiPlaySong(const char* filename)
{
	strcpy(_midi.start_song, filename);
	_midi.playing = true;
	_midi.stop_song = false;
}

extern "C" void ExtMidiStopSong(void)
{
	if (_midi.playing) 
	{
		_midi.stop_song = true;
		_midi.start_song[0] = '\0';
	}
}

extern "C" bool ExtMidiIsPlaying(void)
{
	return _midi.playing;
}

extern "C" void ExtMidiSetVolume(byte vol)
{
	_midi.new_vol = vol;
}

static DWORD WINAPI ExtMidiGo(LPVOID arg)
{
	do {
		char *s;
		s = _midi.start_song;
		if (s[0] != '\0') 
		{
			_TCHAR strWide[260]; // 260 is MAX_PATH
			MULTI_TO_WIDE(s, strWide, sizeof(strWide) );
			_midi.playing = play_midi_file(strWide);
			s[0] = '\0';

			// Delay somewhat in case we don't manage to play.
			if (!_midi.playing) 
			{
				// MYCODE EDIT 5000 to 1000
				Sleep(1000);
			}
		}
		else
		{
			Sleep(1000);
		}
		
		if (_midi.stop_song && _midi.playing) 
		{
			_midi.stop_song = false;
			_midi.playing = false;
			ExtMidiStopSong();
		}

		if (_midi.playing && !ExtMidiIsPlaying())
			_midi.playing = false;

		//WaitForMultipleObjects(1, &_midi.wait_obj, FALSE, 1000);
	} while (!_midi.terminate);

	//DeleteObject(_midi.wait_obj);

	return 0;
}

extern "C" const char* ExtMidiStart(const char* const * parm)
{
	int songCount=0;
	int	got_a_configuration=0;
	int32 output_rate = DEFAULT_RATE;	
	int buffer_fragments = -1;
	int evil = 0;

	memset(&_midi, 0, sizeof(_midi));
	_midi.new_vol = -1;

	song[0] = '\0';	

	if (!read_config_file(CONFIG_FILE)) 
	{
		got_a_configuration=1;
	}
		
	//output_rate = 11025;
	antialiasing_allowed=1;
	buffer_fragments = 3;		/* On Win32 2 is chunky */

	if (!got_a_configuration)
    {
		MessageBox(GetActiveWindow(), TEXT("Cannot read timidity.cfg!"), TEXT("Error!"), MB_ICONSTOP);
		//assert(0);
		//exit(1);
    }
	
	/* Set play mode parameters */
	play_mode->rate = output_rate;
	if (buffer_fragments != -1)
		play_mode->extra_param[0] = buffer_fragments;
	
	
	init_tables ();
		
	/* Open output device */
	if (play_mode->open_output() < 0)
    {
		myprintf (L"Error!", L"Couldn't open %s\n", play_mode->id_name);
		//ctl.close ();
		MessageBox(GetActiveWindow(), TEXT("Cannot open audio device!"), TEXT("Error!"), MB_ICONSTOP);
		//assert(0);
		//exit(1);
    }
		
	if (!control_ratio)
    {
		control_ratio = play_mode->rate / CONTROLS_PER_SECOND;
		if (control_ratio < 1)
			control_ratio = 1;
		else if (control_ratio > MAX_CONTROL_RATIO)
			control_ratio = MAX_CONTROL_RATIO;
    }

	if (*def_instr_name)
	{
		set_default_instrument (def_instr_name);
	}
	
	InitializeCriticalSection (&critSect);
		
	CreateThread(NULL, 8192, ExtMidiGo, 0, 0, &threadId);

	return NULL;
}

#endif /* __MORPHOS__ */
#endif /* __BEOS__ */
