/* $Id: win32.c 3309 2005-12-16 22:01:38Z peter1138 $ */

#include "stdafx.h"
#include "openttd.h"
#include "debug.h"
#include "functions.h"
#include "hal.h"
#include "macros.h"
#include "saveload.h"
#include "ttdstring.h"
#include "table/strings.h"
#include "gfx.h"
#include "window.h"
#include <windows.h>
#include <winnt.h>
#include <wininet.h>
//#include <io.h>
#include "./WINCE/fcntl.h"
#include "variables.h"
#include "win32.h"
#include <ctype.h>

#include "WINCE/UIHelper.h"

bool hires;

static bool _has_console;
static bool cursor_visible = true;

bool MyShowCursor(bool show)
{
	if (cursor_visible == show) return show;

	cursor_visible = show;
	ShowCursor(show);

	return !show;
}

// Helper function needed by dynamically loading SDL
bool LoadLibraryList(Function proc[], const char* dll)
{
	while (*dll != '\0') {
		_TCHAR strWide[MAX_WIDE_STR];
		HMODULE lib;
		MULTI_TO_WIDE( dll, strWide, sizeof(strWide) );
		lib = LoadLibrary(strWide);
		
		if (lib == NULL) return false;
		while (true) {
		  	FARPROC p;

			while (*dll++ != '\0');
			if (*dll == '\0') break;
			MULTI_TO_WIDE( dll, strWide, sizeof(strWide) );
			p = GetProcAddress(lib, strWide);
			if (p == NULL) return false;
			*proc++ = (Function)p;
		}
		dll++;
	}
	return true;
}

void ShowOSErrorBox(const char *buf)
{
	_TCHAR strWide[MAX_WIDE_STR];
	// EDIT: MyShowCursor(true);

	MULTI_TO_WIDE( buf, strWide, sizeof(strWide) );
	MessageBox(GetActiveWindow(), strWide, TEXT("Error!"), MB_ICONSTOP);
}

static inline int strcasecmp(const char* s1, const char* s2)
{
	return _stricmp(s1, s2);
}

static char *_fios_path;
static char *_fios_save_path;
static char *_fios_scn_path;
static FiosItem *_fios_items;
static int _fios_count, _fios_alloc;

static FiosItem *FiosAlloc(void)
{
	if (_fios_count == _fios_alloc) {
		_fios_alloc += 256;
		_fios_items = realloc(_fios_items, _fios_alloc * sizeof(FiosItem));
	}
	return &_fios_items[_fios_count++];
}

static HANDLE MyFindFirstFile(const char *path, const char *file, WIN32_FIND_DATA *fd)
{
	HANDLE h;
	char paths[MAX_PATH];
	_TCHAR strWide[MAX_WIDE_STR];
	const char *s = strrchr(path, '\\');
	
	sprintf(paths, "%s%s%s", path, (s[1] == '\0') ? "" : "\\", file);
	MULTI_TO_WIDE( paths, strWide, sizeof(strWide) );
	h = FindFirstFile(strWide, fd);

	return h;
}

int CDECL compare_FiosItems(const void *a, const void *b)
{
	const FiosItem *da = (const FiosItem *)a;
	const FiosItem *db = (const FiosItem *)b;
	int r;

	if (_savegame_sort_order & SORT_BY_NAME) {
		r = strcasecmp(da->title, db->title);
	} else {
		r = da->mtime < db->mtime ? -1 : 1;
	}

	if (_savegame_sort_order & SORT_DESCENDING) r = -r;
	return r;
}

// Get a list of savegames
FiosItem *FiosGetSavegameList(int *num, int mode)
{
	WIN32_FIND_DATA fd;
	HANDLE h;
	FiosItem *fios;
	int sort_start;
	char strMB[MAX_MB_STR];
	
	if (_fios_save_path == NULL) {
		_fios_save_path = malloc(MAX_PATH);
		strcpy(_fios_save_path, _path.save_dir);
	}

	_fios_path = _fios_save_path;

	// Parent directory, only if not of the type C:\.
	if (_fios_path[3] != '\0') {
		fios = FiosAlloc();
		fios->type = FIOS_TYPE_PARENT;
		fios->mtime = 0;
		strcpy(fios->name, "..");
		strcpy(fios->title, ".. (Parent directory)");
	}

	// Show subdirectories first
	h = MyFindFirstFile(_fios_path, "*.*", &fd);
	if (h != INVALID_HANDLE_VALUE) {
		do 
		{
			WIDE_TO_MULTI( fd.cFileName, strMB, sizeof(strMB) );
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY &&
					strcmp(strMB, ".") != 0 &&
					strcmp(strMB, "..") != 0) {
				fios = FiosAlloc();
				fios->type = FIOS_TYPE_DIR;
				fios->mtime = 0;
				
				ttd_strlcpy(fios->name, strMB, lengthof(fios->name));
				snprintf(fios->title, lengthof(fios->title),
					"%s\\ (Directory)", strMB);
			}
		} while (FindNextFile(h, &fd));
		FindClose(h);
	}

	{
		/* XXX ugly global variables ... */
		byte order = _savegame_sort_order;
		_savegame_sort_order = SORT_BY_NAME | SORT_ASCENDING;
		qsort(_fios_items, _fios_count, sizeof(FiosItem), compare_FiosItems);
		_savegame_sort_order = order;
	}

	// this is where to start sorting
	sort_start = _fios_count;

	/* Show savegame files
	 * .SAV OpenTTD saved game
	 * .SS1 Transport Tycoon Deluxe preset game
	 * .SV1 Transport Tycoon Deluxe (Patch) saved game
	 * .SV2 Transport Tycoon Deluxe (Patch) saved 2-player game
	 */
	h = MyFindFirstFile(_fios_path, "*.*", &fd);
	if (h != INVALID_HANDLE_VALUE) {
		do 
		{
			char *t;

			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) continue;
			
			WIDE_TO_MULTI( fd.cFileName, strMB, sizeof(strMB) );
			t = strrchr(strMB, '.');
			if (t == NULL) continue;

			if (strcasecmp(t, ".sav") == 0) { // OpenTTD
				fios = FiosAlloc();
				fios->type = FIOS_TYPE_FILE;
				fios->mtime = *(uint64*)&fd.ftLastWriteTime;
				ttd_strlcpy(fios->name, strMB, lengthof(fios->name));

				*t = '\0'; // strip extension
				ttd_strlcpy(fios->title, strMB, lengthof(fios->title));

			} else if (mode == SLD_LOAD_GAME || mode == SLD_LOAD_SCENARIO) {
				if (strcasecmp(t, ".ss1") == 0 ||
						strcasecmp(t, ".sv1") == 0 ||
						strcasecmp(t, ".sv2") == 0) { // TTDLX(Patch)
					char buf[MAX_PATH];

					fios = FiosAlloc();
					fios->type = FIOS_TYPE_OLDFILE;
					fios->mtime = *(uint64*)&fd.ftLastWriteTime;
					ttd_strlcpy(fios->name, strMB, lengthof(fios->name));
					sprintf(buf, "%s\\%s", _fios_path, strMB);
					GetOldSaveGameName(fios->title, buf);
				}
			}
		} while (FindNextFile(h, &fd));
		FindClose(h);
	}

	qsort(_fios_items + sort_start, _fios_count - sort_start, sizeof(FiosItem), compare_FiosItems);

	// Drives
	{
		fios = FiosAlloc();
		fios->type = FIOS_TYPE_DRIVE;
		sprintf(fios->name, "\\");
		sprintf(fios->title, "\\");
	}

	*num = _fios_count;
	return _fios_items;
}

// Get a list of scenarios
FiosItem *FiosGetScenarioList(int *num, int mode)
{
	FiosItem *fios;
	WIN32_FIND_DATA fd;
	HANDLE h;
	int sort_start;
	char strMB[MAX_MB_STR];
	
	if (_fios_scn_path == NULL) {
		_fios_scn_path = malloc(MAX_PATH);
		strcpy(_fios_scn_path, _path.scenario_dir);
	}

	_fios_path = _fios_scn_path;

	// Parent directory, only if not of the type C:\.
	if (_fios_path[3] != '\0' && mode != SLD_NEW_GAME) {
		fios = FiosAlloc();
		fios->type = FIOS_TYPE_PARENT;
		fios->mtime = 0;
		strcpy(fios->title, ".. (Parent directory)");
	}

	// Show subdirectories first
	h = MyFindFirstFile(_fios_scn_path, "*.*", &fd);
	if (h != INVALID_HANDLE_VALUE && mode != SLD_NEW_GAME) {
		do {
			WIDE_TO_MULTI( fd.cFileName, strMB, sizeof(strMB) );
			if (fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY &&
					strcmp(strMB, ".") != 0 &&
					strcmp(strMB, "..") != 0) {
				fios = FiosAlloc();
				fios->type = FIOS_TYPE_DIR;
				fios->mtime = 0;
				ttd_strlcpy(fios->name, strMB, lengthof(fios->name));
				snprintf(fios->title, lengthof(fios->title),
					"%s\\ (Directory)", strMB);
			}
		} while (FindNextFile(h, &fd));
		FindClose(h);
	}

	{
		/* XXX ugly global variables ... */
		byte order = _savegame_sort_order;
		_savegame_sort_order = SORT_BY_NAME | SORT_ASCENDING;
		qsort(_fios_items, _fios_count, sizeof(FiosItem), compare_FiosItems);
		_savegame_sort_order = order;
	}

	// this is where to start sorting
	sort_start = _fios_count;

	/* Show scenario files
	 * .SCN OpenTTD style scenario file
	 * .SV0 Transport Tycoon Deluxe (Patch) scenario
	 * .SS0 Transport Tycoon Deluxe preset scenario
	 */
	h = MyFindFirstFile(_fios_scn_path, "*.*", &fd);
	if (h != INVALID_HANDLE_VALUE) {
		do {
			char *t;

			if ((fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) continue;
			
			WIDE_TO_MULTI( fd.cFileName, strMB, sizeof(strMB) );
			t = strrchr(strMB, '.');
			if (t == NULL) continue;

			if (strcasecmp(t, ".scn") == 0) { // OpenTTD
				fios = FiosAlloc();
				fios->type = FIOS_TYPE_SCENARIO;
				fios->mtime = *(uint64*)&fd.ftLastWriteTime;
				ttd_strlcpy(fios->name, strMB, lengthof(fios->name));

				*t  = '\0'; // strip extension
				ttd_strlcpy(fios->title, strMB, lengthof(fios->title));
			} else if (mode == SLD_LOAD_GAME || mode == SLD_LOAD_SCENARIO ||
					mode == SLD_NEW_GAME) {
				if (strcasecmp(t, ".sv0") == 0 ||
						strcasecmp(t, ".ss0") == 0) { // TTDLX(Patch)
					char buf[MAX_PATH];

					fios = FiosAlloc();
					fios->type = FIOS_TYPE_OLD_SCENARIO;
					fios->mtime = *(uint64*)&fd.ftLastWriteTime;
					sprintf(buf, "%s\\%s", _fios_path, strMB);
					GetOldScenarioGameName(fios->title, buf);
					ttd_strlcpy(fios->name, strMB, lengthof(fios->name));
				}
			}
		} while (FindNextFile(h, &fd));
		FindClose(h);
	}

	qsort(_fios_items + sort_start, _fios_count - sort_start, sizeof(FiosItem), compare_FiosItems);

	// Drives
	{
		fios = FiosAlloc();
		fios->type = FIOS_TYPE_DRIVE;
		sprintf(fios->name, "\\");
		sprintf(fios->title, "\\");
	}


	*num = _fios_count;
	return _fios_items;
}


// Free the list of savegames
void FiosFreeSavegameList(void)
{
	free(_fios_items);
	_fios_items = NULL;
	_fios_alloc = _fios_count = 0;
}

// Browse to
char *FiosBrowseTo(const FiosItem *item)
{
	char *path = _fios_path;
	char *s;

	switch (item->type) {
		case FIOS_TYPE_DRIVE:
			sprintf(path, "%c:\\", item->title[0]);
			break;

		case FIOS_TYPE_PARENT:
			s = strrchr(path, '\\');
			if (s != path + 2)
				s[0] = '\0';
			else
				s[1] = '\0';
			break;

		case FIOS_TYPE_DIR:
			if (path[3] != '\0') strcat(path, "\\");
			strcat(path, item->name);
			break;

		case FIOS_TYPE_DIRECT:
			sprintf(path, "%s\\", item->name);
			s = strrchr(path, '\\');
			if (s[1] == '\0') s[0] = '\0'; // strip trailing slash
			break;

		case FIOS_TYPE_FILE:
		case FIOS_TYPE_OLDFILE:
		case FIOS_TYPE_SCENARIO:
		case FIOS_TYPE_OLD_SCENARIO: {
			static char str_buffr[512];

			sprintf(str_buffr, "%s\\%s", path, item->name);
			return str_buffr;
		}
	}

	return NULL;
}

/**
 * Get descriptive texts. Returns the path and free space
 * left on the device
 * @param path string describing the path
 * @param tfs total free space in megabytes, optional (can be NULL)
 * @return StringID describing the path (free space or failure)
 */
StringID FiosGetDescText(const char **path, uint32 *tot)
{
	StringID sid;
	*path = _fios_path;
	*tot = 99000000 >> 20;
	sid = STR_4005_BYTES_FREE;
	return sid;
}

void FiosMakeSavegameName(char *buf, const char *name, size_t size)
{
	const char* extension;
	const char* period;

	if (_game_mode == GM_EDITOR)
		extension = ".scn";
	else
		extension = ".sav";

	// Don't append the extension, if it is already there
	period = strrchr(name, '.');
	if (period != NULL && strcasecmp(period, extension) == 0) extension = "";

	snprintf(buf, size, "%s\\%s%s", _fios_path, name, extension);
}

bool FiosDelete(const char *name)
{
	char path[512];
	_TCHAR strWide[MAX_WIDE_STR];

	snprintf(path, lengthof(path), "%s\\%s", _fios_path, name);
	MULTI_TO_WIDE( path, strWide, sizeof(strWide) );
	return DeleteFile(strWide) != 0;
}

bool FileExists(const char *filename)
{
	HANDLE hand;
	_TCHAR strWide[MAX_WIDE_STR];

	MULTI_TO_WIDE( filename, strWide, sizeof(strWide) );
	hand = CreateFile(strWide, 0, 0, NULL, OPEN_EXISTING, 0, NULL);
	if (hand == INVALID_HANDLE_VALUE) return false;
	CloseHandle(hand);
	return true;
}

static int CDECL LanguageCompareFunc(const void *a, const void *b)
{
	return strcmp(*(const char* const *)a, *(const char* const *)b);
}

int GetLanguageList(char **languages, int max)
{
	HANDLE hand;
	int num = 0;
	char filedir[MAX_PATH];
	WIN32_FIND_DATA fd;
	_TCHAR strWide[MAX_WIDE_STR];
	char strMB[MAX_MB_STR];
	sprintf(filedir, "%s*.lng", _path.lang_dir);

	MULTI_TO_WIDE( filedir, strWide, sizeof(strWide) );
	hand = FindFirstFile(strWide, &fd);
	if (hand != INVALID_HANDLE_VALUE) {
		do {
			if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) 
			{
				WIDE_TO_MULTI( fd.cFileName, strMB, sizeof(strMB) );
				languages[num++] = _strdup(strMB);
				if (num == max) break;
			}
		} while (FindNextFile(hand, &fd));
		FindClose(hand);
	}

	qsort(languages, num, sizeof(char*), LanguageCompareFunc);
	return num;
}

static int ParseCommandLine(char *line, char **argv, int max_argc)
{
	int n = 0;

	do {
		// skip whitespace
		while (*line == ' ' || *line == '\t')
			line++;

		// end?
		if (*line == '\0')
			break;

		// special handling when quoted
		if (*line == '"') {
			argv[n++] = ++line;
			while (*line != '"') {
				if (*line == '\0') return n;
				line++;
			}
		} else {
			argv[n++] = line;
			while (*line != ' ' && *line != '\t') {
				if (*line == '\0') return n;
				line++;
			}
		}
		*line++ = '\0';
	} while (n != max_argc);

	return n;
}

void ShowInfo(const char *str)
{
/*
	if (_has_console)
		puts(str);
	else {
		bool old;

		ReleaseCapture();
		_left_button_clicked =_left_button_down = false;

		old = MyShowCursor(true);
		if (MessageBoxA(GetActiveWindow(), str, "OpenTTD", MB_ICONINFORMATION | MB_OKCANCEL) == IDCANCEL) {
			CreateConsole();
		}
		MyShowCursor(old);
	}
*/
	FILE *f;
	f = _tfopen(TEXT("\\debug.txt"), TEXT("a"));
	if (f == NULL) return;
	fprintf(f, "%s\r\n", str);

	fclose(f);
}

char *GetPath()
{
        static _TCHAR lpFilename[MAX_PATH + 1]; // fix memory leak
        _TCHAR *pDest; 
        int nStrLen; 
        int result; 
        int ch = '\\'; 
		static char strMB[MAX_PATH + 1];
        nStrLen = GetModuleFileName(NULL, lpFilename, MAX_PATH + 1); 

        pDest = wcsrchr(lpFilename, ch); 
		
        if (pDest != NULL) 
        { 
			result = pDest - lpFilename + 1;
			lpFilename[result] = '\0';
			WIDE_TO_MULTI( lpFilename, strMB, sizeof(strMB) );
			return strMB; 
        } 
        else 
        {
			error("GetPath failed");
			return NULL; 
        }
}

void DeterminePaths(void)
{
	char *cfg;
	_TCHAR strWide[MAX_WIDE_STR];

	_path.personal_dir = _path.game_data_dir = cfg = malloc(MAX_PATH);
	// MYCODE EDIT
	cfg = GetPath();
	_path.personal_dir = cfg;
	_path.game_data_dir = cfg;

	_path.save_dir = str_fmt("%ssave", cfg);
	_path.autosave_dir = str_fmt("%s\\autosave", _path.save_dir);
	_path.scenario_dir = str_fmt("%sscenario", cfg);
	_path.gm_dir = str_fmt("%sgm\\", cfg);
	_path.data_dir = str_fmt("%sdata\\", cfg);
	_path.lang_dir = str_fmt("%slang\\", cfg);

	if (_config_file == NULL)
		_config_file = str_fmt("%sopenttd.cfg", _path.personal_dir);

	_highscore_file = str_fmt("%shs.dat", _path.personal_dir);
	_log_file = str_fmt("%sopenttd.log", _path.personal_dir);

	// make (auto)save and scenario folder
	MULTI_TO_WIDE( _path.save_dir, strWide, sizeof(strWide) );
	CreateDirectory(strWide, NULL);
	MULTI_TO_WIDE( _path.autosave_dir, strWide, sizeof(strWide) );
	CreateDirectory(strWide, NULL);
	MULTI_TO_WIDE( _path.scenario_dir, strWide, sizeof(strWide) );
	CreateDirectory(strWide, NULL);
}

int CDECL snprintf(char *str, size_t size, const char *format, ...)
{
	va_list ap;
	int ret;

	va_start(ap, format);
	ret = vsnprintf(str, size, format, ap);
	va_end(ap);
	return ret;
}

#ifdef _MSC_VER
#if _MSC_VER < 1400 /* Already defined in VS 2005 */
int CDECL vsnprintf(char *str, size_t size, const char *format, va_list ap)
{
	int ret;
	ret = _vsnprintf(str, size, format, ap);
	if (ret < 0) str[size - 1] = '\0';
	return ret;
}
#endif
#endif

/**
 * Insert a chunk of text from the clipboard onto the textbuffer. Get TEXT clipboard
 * and append this up to the maximum length (either absolute or screenlength). If maxlength
 * is zero, we don't care about the screenlength but only about the physical length of the string
 * @param tb @Textbuf type to be changed
 * @return Return true on successfull change of Textbuf, or false otherwise
 */
bool InsertTextBufferClipboard(Textbuf *tb)
{
	if (IsClipboardFormatAvailable(CF_TEXT)) {
		HGLOBAL cbuf;
		const byte *data, *dataptr;
		uint16 width = 0;
		uint16 length = 0;

		OpenClipboard(NULL);
		cbuf = GetClipboardData(CF_TEXT);
		data = GlobalLock(cbuf); // clipboard data
		dataptr = data;

		for (; IsValidAsciiChar(*dataptr) && (tb->length + length) < tb->maxlength - 1 &&
				(tb->maxwidth == 0 || width + tb->width + GetCharacterWidth((byte)*dataptr) <= tb->maxwidth); dataptr++) {
					width += GetCharacterWidth((byte)*dataptr);
			length++;
		}

		if (length == 0)
			return false;

		memmove(tb->buf + tb->caretpos + length, tb->buf + tb->caretpos, tb->length - tb->caretpos);
		memcpy(tb->buf + tb->caretpos, data, length);
		tb->width += width;
		tb->caretxoffs += width;

		tb->length += length;
		tb->caretpos += length;
		tb->buf[tb->length + 1] = '\0'; // terminating zero

		GlobalUnlock(cbuf);
		CloseClipboard();
		return true;
	}
	return false;
}


void CSleep(int milliseconds)
{
	Sleep(milliseconds);
}


// Utility function to get the current timestamp in milliseconds
// Useful for profiling
int64 GetTS(void)
{
	static double freq;
	__int64 value;
	if (!freq) {
		QueryPerformanceFrequency((LARGE_INTEGER*)&value);
		freq = (double)1000000 / value;
	}
	QueryPerformanceCounter((LARGE_INTEGER*)&value);
	return (__int64)(value * freq);
}

int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	LPTSTR lpCmdLine, int nCmdShow)
{
	hires = false;

	HIDPI_InitScaling();
	
	if( g_HIDPI_Width == 480 || g_HIDPI_Width == 640 )
	{
		hires = true;
	}
	
	ttd_main();

	return 0;
}
