/*

    TiMidity -- Experimental MIDI to WAVE converter
    Copyright (C) 1995 Tuukka Toivonen <toivonen@clinet.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
    
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    
    win_audio.c
    
    Functions to play sound on the Win32 audio driver (Win 95 or Win NT).

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "config.h"
#include "output.h"

static int open_output (void);	/* 0=success, 1=warning, -1=fatal error */
static void close_output (void);
static void output_data (int32b * buf, int32b count);
static void flush_output (void);
static void purge_output (void);

/* export the playback mode */

#define dpm symbian_play_mode

PlayMode dpm = {
	DEFAULT_RATE,
	PE_16BIT | PE_SIGNED,
	-1,
	{16},
	"Symbian audio driver",
	'd',
	NULL,
	open_output,
	close_output,
	output_data,
	flush_output,
	purge_output
};

/* Max audio blocks waiting to be played */
static int nBlocks;

//extern CRITICAL_SECTION critSect;

static void
wait (void)
{
	while (nBlocks>1)
		User::AfterHighRes (1000);
}

extern "C" 	void add_block(char* aDataBuffer, int aLen);

extern "C" void block_played()
{
	nBlocks--;
}


static int
open_output (void)
{
	int  mono, eight_bit, warnings = 0;
	
	// They can't mean these 
	dpm.encoding &= ~(PE_ULAW | PE_BYTESWAP);
	
	if (dpm.encoding & PE_16BIT)
		dpm.encoding |= PE_SIGNED;
	else
		dpm.encoding &= ~PE_SIGNED;
	
	mono = (dpm.encoding & PE_MONO);
	eight_bit = !(dpm.encoding & PE_16BIT);

	nBlocks = 0;
	return warnings;
}

static void
output_data (int32b * buf, int32b count)
{
	int len = count;
	int waitcnt = 100;
	if (!(dpm.encoding & PE_MONO))	// Stereo sample 
    {
		count *= 2;
		len *= 2;
    }
	
	if (dpm.encoding & PE_16BIT)
		len *= 2;
	
	if (dpm.encoding & PE_16BIT)
		// Convert data to signed 16-bit PCM 
		s32tos16 (buf, count);
	else
		// Convert to 8-bit unsigned. 
		s32tou8 (buf, count);
	
	while (nBlocks >= dpm.extra_param[0] && waitcnt>0)
	{
		waitcnt--;
		User::AfterHighRes (1000);
	}

	if(nBlocks < dpm.extra_param[0])
	{
	add_block((char*) buf, len);

	nBlocks++;	
	}
}

static void
close_output (void)
{
}

static void
flush_output (void)
{
	wait ();
}

static void
purge_output (void)
{
	//waveOutReset (dev);
	wait ();
}
